/* 
 * Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
 *
 * This file is part of Rast.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 *
 * Linking the Rast statically or dynamically with other modules is
 * making a combined work based on the Rast.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of the Rast give
 * you permission to link the Rast with independent modules that
 * communicate with the Rast solely through the the Apache Portable
 * Runtime interface, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * combined work under terms of your choice, provided that every
 * copy of the combined work is accompanied by a complete copy of
 * the source code of the Rast (the version of the Rast used to
 * produce the combined work), being distributed under the terms of
 * the GNU General Public License plus this exception.  An
 * independent module is a module which is not derived from or based
 * on the Rast.
 *
 * Note that people who make modified versions of the Rast are not
 * obligated to grant this special exception for their modified
 * versions; it is their choice whether to do so.  The GNU General
 * Public License gives permission to release a modified version
 * without this exception; this exception also makes it possible to
 * release a modified version which carries forward this exception.
 */

#include <apr_strings.h>
#include <ruby.h>
#include <st.h>

#undef PACKAGE_NAME
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef PACKAGE_STRING
#include "rast/config.h"
#include "rast/xmlrpc_client.h"

static VALUE eError, eRastError, eAprError, eBDBError;
static VALUE cPool;

static void
pool_free(apr_pool_t *pool)
{
    if (pool) {
        apr_pool_destroy(pool);
    }
}

static VALUE
pool_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, pool_free, NULL);
}

static int
raise_on_pool_failure(int retcode)
{
    rb_memerror();
    return -1;  /* prevent compiler warnings */
}

static VALUE
pool_initialize(VALUE self)
{
    apr_status_t status;
    apr_pool_t *pool;

    status = apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    if (status != APR_SUCCESS) {
        rb_memerror();
    }
    DATA_PTR(self) = pool;
    return Qnil;
}

static VALUE
pool_new()
{
    VALUE obj;

    obj = pool_alloc(cPool);
    pool_initialize(obj);
    return obj;
}

static apr_pool_t *
get_pool(VALUE self)
{
    if (TYPE(self) != T_DATA ||
        RDATA(self)->dfree != (RUBY_DATA_FUNC) pool_free) {
        rb_raise(rb_eTypeError,
                 "wrong argument type %s (expected Rast::Pool)",
                 rb_obj_classname(self));
    }
    return (apr_pool_t *) DATA_PTR(self);
}

static void
raise_error(rast_error_t *error)
{
    VALUE error_class;
    char buf[RAST_BUFSIZ];

    switch (error->type) {
    case RAST_ERROR_TYPE_RAST:
        error_class = eRastError;
        break;
    case RAST_ERROR_TYPE_APR:
        error_class = eAprError;
        break;
    case RAST_ERROR_TYPE_BDB:
        error_class = eBDBError;
        break;
    default:
        error_class = eError;
    };

#ifdef RAST_DEBUG
    snprintf(buf, RAST_BUFSIZ, "%s:%d: %s",
             error->file, error->line, error->message);
#else
    strncpy(buf, error->message, RAST_BUFSIZ);
#endif
    buf[RAST_BUFSIZ - 1] = '\0';
    rast_error_destroy(error);
    rb_raise(error_class, "%s", buf);
}

typedef struct {
    rast_db_t *db;
    apr_pool_t *pool;
    int closed;
} db_data_t;

static void
get_int_option(VALUE options, const char *name, int *dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (RTEST(value)) {
        *dst = NUM2INT(value);
    }
}

static VALUE
process_db_initialize(int argc, VALUE *argv, VALUE self,
                      rast_error_t *(*open)(rast_db_t **db, const char *name,
                                            int flags,
                                            rast_db_open_option_t *options,
                                            apr_pool_t *pool))
{
    VALUE name, vflags, voptions, vpool;
    rast_error_t *error;
    apr_pool_t *pool, *tmp_pool;
    db_data_t *data;
    rast_db_t *db;
    int flags = RAST_RDWR;
    rast_db_open_option_t *options = NULL;

    rb_scan_args(argc, argv, "12", &name, &vflags, &voptions);
    if (!NIL_P(vflags)) {
        flags = NUM2INT(vflags);
    }
    if (!NIL_P(voptions)) {
        vpool = pool_new();
        tmp_pool = get_pool(vpool);
        options = rast_db_open_option_create(tmp_pool);
        get_int_option(voptions, "sync_threshold_chars",
                       &options->sync_threshold_chars);
    }
    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    SafeStringValue(name);
    error = open(&db, RSTRING(name)->ptr, flags, options, pool);
    if (error != RAST_OK) {
        apr_pool_destroy(pool);
        raise_error(error);
    }
    data = ALLOC(db_data_t);
    data->db = db;
    data->pool = pool;
    data->closed = 0;
    DATA_PTR(self) = data;
    return Qnil;
}

static VALUE
xmlrpc_client_initialize(int argc, VALUE *argv, VALUE self)
{
    return process_db_initialize(argc, argv, self, rast_xmlrpc_client_open);
}

void
Init_rast_xmlrpc_client()
{
    VALUE mRast, cDB, cXMLRPCClient;

    rb_require("rast");

    rast_xmlrpc_client_initialize();
    atexit(rast_xmlrpc_client_finalize);

    mRast = rb_const_get(rb_mKernel, rb_intern("Rast"));
    eError = rb_const_get(mRast, rb_intern("Error"));
    eRastError = rb_const_get(mRast, rb_intern("RastError"));
    eAprError = rb_const_get(mRast, rb_intern("AprError"));
    eBDBError = rb_const_get(mRast, rb_intern("BDBError"));

    cDB = rb_const_get(mRast, rb_intern("DB"));
    cPool = rb_const_get(mRast, rb_intern("Pool"));

    cXMLRPCClient = rb_define_class_under(mRast, "XMLRPCClient", cDB);
    rb_define_method(cXMLRPCClient, "initialize",
                     xmlrpc_client_initialize, -1);
}

/* vim: set filetype=c sw=4 expandtab : */
